# =============================================================================
# ~/spinnaker/associate/neal3way.py
#
# created January 2020 Lungsi
#
# =============================================================================
from types import SimpleNamespace as structdata

import spynnaker8 as sim

import quantities as pq
# for plotting
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import gridspec
from matplotlib import cm

from .nealassoc.readInheritanceFile import InheritanceReaderClass
from .nealassoc.readUnitFile import UnitReaderClass
#from readAssocFile import AssocReaderClass
from .nealassoc.make3Assoc import NeuralThreeAssocClass
from nmcog.spinnaker.specialfunction.neal import NealCoverFunctions

class NEAL3Way(object):
    """Neuromorphic Embodied Agents that Learn `(NEAL) <http://www.cwa.mdx.ac.uk/NEAL/NEAL.html>`_ based three-way association function.
    
    Example
    
    ::
    
        bases = {"units": ["animal", "mammal", "bird", "canary"],
                 "is-a": [ ["canary", "bird"], ["bird", "animal"], ["mammal", "animal"] ]}
        associate = {"properties": ["food", "fur", "flying", "yellow"], # properties to be associated between base units and its relations
                     "relations": ["eats", "likes", "travels", "has", "colored"], # relations associated with properties and base units
                     "connections": [ ["animal", "eats", "food"], ["mammal", "has", "fur"], # specific combos of base-props-relations
                                      ["bird", "travels", "flying"], ["canary", "colored", "yellow"]] }
    
    **Note:**
    
    * One may think of the arguments, ``bases`` and ``associate`` as some a priori knowledge of the embodied agent.
    * The value for "is-a" must have **at least one** non-empty list.
    * The result of above restriction is that ``bases`` must have at least two units, i.e. list of the value for "units" have at least two elements.
    * Usually, the number of "properties" equal the number of base units.
    * The number of "relations" can be of any number as long as there is at least one.
    * Finally, the components within "connections" is of the form ``[<base-unit>, <a-relation>, <a-property>]``.
    * Looking from the context of a sentence one may view ``bases`` as the subject, "properties" (key of ``associate``) as the object and "relations" (also key of ``associate``) as the predicate, and "connections" as the sentence structure.
    * Specifically the instance of :ref:`InheritanceReaderClass` is the base data for :ref:`NEAL3Way`.
    
    
    Considering the above example whose base data is the instance such that,
        
    ::
    
        >>> print(basedata.units)
        ["animal", "mammal", "bird", "canary"]
        
           
    Since there are four association units in the base data, four cell assemblies will be created.
    
    ::
    
             CA for animal            CA for mammal              CA for bird              CA for canary
        ooooooooooooooooooooooo   ooooooooooooooooooooooo   ooooooooooooooooooooooo   ooooooooooooooooooooooo
        o                     o   o                     o   o                     o   o                     o
        o (0) (2) (4) (6) (8) o   o (0) (2) (4) (6) (8) o   o (0) (2) (4) (6) (8) o   o (0) (2) (4) (6) (8) o
        o                     o   o                     o   o                     o   o                     o
        o (1) (3) (5) (7) (9) o   o (1) (3) (5) (7) (9) o   o (1) (3) (5) (7) (9) o   o (1) (3) (5) (7) (9) o
        o                     o   o                     o   o                     o   o                     o
        ooooooooooooooooooooooo   ooooooooooooooooooooooo   ooooooooooooooooooooooo   ooooooooooooooooooooooo
    
    
    * Refer to :ref:`FSAHelperFunctions` ``.makeCA`` about the structure of a cell assembly based on the Neuromorphic Embodied Agents that Learn `(NEAL) <http://www.cwa.mdx.ac.uk/NEAL/NEAL.html>`_.
    * Refer to :ref:`FSAHelperFunctions` ``.getCAConnectors`` for details on how the cell units in the `neuron population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a cell assembly are connected.
    
    Connection (excitatory) between the cell assemblies is based on the "isA" relationship pair.
    
    ::
    
        >>> print(basedata.isARelationships)
        [ ["canary", "bird"], ["bird", "animal"], ["mammal", "animal"] ]
    
    
    Therefore a hierarchy topology will be created
    
    ::
    
             ;;;;;;;;;;;;;;;;;;;;;;;;;;              ;;;;;;;;;;;;;;;;;;;;;;;;;;
             ;;;;;;;;;;;;;;;;;;;;;;;;;;              ;;;;;;;;;;;;;;;;;;;;;;;;;;
           ..;;;;;..              ;;;;;            ..;;;;;..              ;;;;;
             ':::'                ;;;;;              ':::'                ;;;;;
              ':'                 ;;;;;               ':'                 ;;;;;
        ooooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo
        o               o   o               o   o               o   o               o
        o CA for animal o   o CA for mammal o   o  CA for bird  o   o CA for canary o
        o               o   o               o   o               o   o               o
        ooooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo
               .                                     ;;;;;
             .:;:.                                   ;;;;;
           .:;;;;;:.                                 ;;;;;
             ;;;;;                                   ;;;;;
             ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
             ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    
    This hierarchy topology is also refered to a the network of the association units in the base data.
    
    * Refer to :ref:`NeuralInheritanceClass` ``.makeHiersFromHier`` for how the excitatory connections among the assembly of association units are made.
    
    Since,
    
    ::
    
        >>> print(propdata.units)
        ["food", "fur", "flying", "yellow"]
        
        >>> print(reldata.units)
        ["eats", "likes", "travels", "has", "colored"]
    
    
    The association topology is
    
    ::
    
        oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo
        o                o   o                o   o                o   o                o
        o  CA for food   o   o   CA for fur   o   o CA for flying  o   o CA for yellow  o
        o                o   o                o   o                o   o                o
        oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo
        
        oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo
        o                o   o                o   o                o   o                o   o                o
        o  CA for eats   o   o  CA for likes  o   o CA for travels o   o   CA for has   o   o CA for colored o
        o                o   o                o   o                o   o                o   o                o
        oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo
    
    
    Given an association tuple (list of three strings representing base, relation, and property),
    
    ::
    
        >>> print(assocdata.assocs)
        [ ["animal", "eats", "food"], ["mammal", "has", "fur"], ["bird", "travels", "flying"], ["canary", "colored", "yellow"] ]
    
    
    * The hierarchy topology to the cell assemblies for property.
    * The hierarchy topology to the cell assemblies for relation.
    * The cell assemblies for property to the hierarchy topology.
    * The cell assemblies for property to the cell assemblies for relation.
    * The cell assemblies for relation to the hierarchy topology.
    * The cell assemblies for relation to the cell assemblies for property.
    
    For illustration below shows only for `[ ["animal", "eats", "food"], ["canary", "colored", "yellow"] ]`
    
    ::
    
                           ;;;;;;;;;;;;;;;;;;;;;;                   ;;;;;;;;;;;;;;;;;;;;;;
                           \/                  ;;                   \/                  ;;
               .    ooooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo    .
        .......;;.  o               o   o               o   o               o   o               o  .;;..............
        ;;;::::;;;;.o CA for animal o   o CA for mammal o   o  CA for bird  o   o CA for canary o.;;;;:::::::::::;;;
        ;;;::::;;:' o               o   o               o   o               o   o               o ':;;:::::::::::;;;
        ;;;    :'   ooooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo   ':           ;;;
        ;;;                /\                                       ;;                                           ;;;
        ;;;                ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                                           ;;;
        ;;;                                                                                                      ;;;
        ;;;    .    oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo    .       ;;;
        ;;;....;;.  o                o   o                o   o                o   o                o  .;;.......;;;
        ;;;::::;;;;.o  CA for food   o   o   CA for fur   o   o CA for flying  o   o CA for yellow  o.;;;;:::::::;;;
        ;;;::::;;:' o                o   o                o   o                o   o                o ':;;::::::;;;;;
        ;;;    :'   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   ':     ':::::'
        ;;;                                                                                                      ':'
        ;;;    .    oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo
        ;;;....;;.  o                o   o                o   o                o   o                o   o                o
        ;;;::::;;;;.o  CA for eats   o   o  CA for likes  o   o CA for travels o   o   CA for has   o   o CA for colored o
        ;;;::::;;:' o                o   o                o   o                o   o                o   o                o
               :'   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo
    
    
    """
    
    def __init__(self, bases, associate, turnon="all", twotest=None):
        # bases = {"units": ["animal", "mammal", "bird", "canary"],
        #          "relations": [ ["canary", "bird"], ["bird", "animal"], ["mammal", "animal"] ]}
        # associate = {"properties": ["food", "fur", "flying", "yellow"], # properties to be associated between base units and its relations
        #              "relations": ["eats", "likes", "travels", "has", "colored"], # relations associated with properties and base units
        #              "connections": [ ["animal", "eats", "food"], ["mammal", "has", "fur"], # specific combos of base-props-relations
        #                               ["bird", "travels", "flying"], ["canary", "colored", "yellow"] ] }
        neal = NealCoverFunctions()
        sim.setup(timestep=neal.DELAY, min_delay=neal.DELAY, max_delay=neal.DELAY, debug=0)
        #
        self.__create_datastructures(bases, associate)
        self.__create_network()
        self.__choose_applicable_test( turnon, twotest )
        #
        neal.nealApplyProjections()
        sim.run(self.simTime)
        [neo_base, neo_property, neo_relation] = self.__getdata()
        self.results = { "base": self.__split_spiketrains("basedata", neo_base, turnon),
                         "property": self.__split_spiketrains("propdata", neo_property, turnon),
                         "relation": self.__split_spiketrains("reldata", neo_relation, turnon) }
        #sim.reset()
        sim.end()
        
    def get_results(self):
        """Returns a dictionary with keys "base", "property", and "relation" whose values are dictionaries.
        A value dictionary is such that the keys are the names of the cell assemblies (i.e. "units" in respective structured data)
        and values are the respective spike trains (i.e. from all cell units in an assembly).
        """
        return self.results
        
    def __create_datastructures(self, bases, associate):
        """."""
        #self.basedata = structdata()
        self.basedata = InheritanceReaderClass()
        self.basedata.numberUnits = len(bases["units"])
        self.basedata.units = bases["units"]
        self.basedata.isARelationships = bases["is-a"]
        #self.basedata.getUnitNumber = lambda checkUnit: [ resultUnit for resultUnit in range(0, self.basedata.numberUnits)
        #                                                              if checkUnit==self.basedata.units[resultUnit] ][0]
        #self.propdata = structdata()
        self.propdata = UnitReaderClass()
        self.propdata.numberUnits = len(associate["properties"])
        self.propdata.units = associate["properties"]
        #self.propdata.getUnitNumber = lambda checkUnit: [ resultUnit for resultUnit in range(0, self.propdata.numberUnits)
        #                                                              if checkUnit==self.propdata.units[resultUnit] ][0]
        #self.reldata = structdata()
        self.reldata = UnitReaderClass()
        self.reldata.numberUnits = len(associate["relations"])
        self.reldata.units = associate["relations"]
        #self.reldata.getUnitNumber = lambda checkUnit: [ resultUnit for resultUnit in range(0, self.reldata.numberUnits)
        #                                                              if checkUnit==self.reldata.units[resultUnit] ][0]
        self.assocdata = structdata() #AssocReaderClass()
        self.assocdata.numberAssocs = len(associate["connections"])
        self.assocdata.assocs = associate["connections"]
        
    def __create_network(self):
        """Moved to main description."""
        self.neural3assoc_topology = NeuralThreeAssocClass()
        self.neural3assoc_topology.createBaseNet( self.basedata )
        self.neural3assoc_topology.createAssociationTopology( self.propdata, self.reldata )
        self.neural3assoc_topology.addAssociations( self.assocdata )
        
    def __choose_applicable_test(self, turnon, twotest=None):
        """If ``turnon = "all"`` then all the cell assemblies for each subject (base), object (property) and relation (predicate) is activated.
        On the other hand, if ``turnon = [<subject-name>, <predicate-name>, <object-name>]`` only cell assemblies for these are trigerred."""
        if turnon=="all":
            self.simTime = self.neural3assoc_topology.createUnitTests() + 100
            self.test_metadata = [turnon, twotest]
        elif (turnon!="all" and twotest=="prime"):
            baseNum = self.__check_and_returnUnitNumber( self.basedata, turnon, 0 ) # base or subject
            probNum = self.__check_and_returnUnitNumber( self.propdata, turnon, 1 ) # property or object
            relNum = self.__check_and_returnUnitNumber( self.reldata, turnon, 2 ) # relation or predicate
            self.neural3assoc_topology.createTwoPrimeTest( baseNum, probNum, relNum )
            self.simTime = 200.0
            self.test_metadata = [None, "prime"]
        elif (turnon!="all" and twotest is None):
            baseNum = self.__check_and_returnUnitNumber( self.basedata, turnon, 0 ) # base or subject
            probNum = self.__check_and_returnUnitNumber( self.propdata, turnon, 1 ) # property or object
            relNum = self.__check_and_returnUnitNumber( self.reldata, turnon, 2 ) # relation or predicate
            self.neural3assoc_topology.createTwoTest( baseNum, probNum, relNum )
            self.simTime = 200.0
            self.test_metadata = [None, None]
    
    # Private function for returning a unit number for respective structured data
    def __check_and_returnUnitNumber(self, strdata, turnon, indx):
        try:
            someNum = strdata.getUnitNumber( turnon[indx] )
        except:
            someNum = -1 # cell assembly will not be activated
        return someNum
    
    # Private function for getting Neo object for all the created cell assemblies.
    def __getdata(self):
        """Gets the recorded `Neo <https://neo.readthedocs.io/en/latest/>`_ objects."""
        neo_base = self.neural3assoc_topology.neuralHierarchyTopology.cells.get_data(variables=["spikes"])
        neo_property = self.neural3assoc_topology.propertyCells.get_data(variables=["spikes"])
        neo_relation = self.neural3assoc_topology.relationCells.get_data(variables=["spikes"])
        return [neo_base, neo_property, neo_relation]
        
    # Private function for return overall spike train 
    def __get_overallspikes(self, dataname, neo_data, turnon):
        """Returns `Neo SpikeTrain <https://neo.readthedocs.io/en/stable/api_reference.html#neo.core.SpikeTrain>`_.
        All of it if the data structure is basedata other wise the they are shifted accordingly for propdata and reldata.
        
        This function is no longer used by __split_spiketrains but left here for potential/future use.
        """
        if turnon=="all":
            if dataname=="basedata":
                return neo_data.segments[0].spiketrains
            else:
                if dataname=="propdata":
                    data = getattr(self, "basedata")
                    tstart = data.numberUnits * 100 # this 100 is the same value as in :py:meth:`.__choose_applicable_test`
                else:
                    data1 = getattr(self, "basedata")
                    data2 = getattr(self, "propdata")
                    tstart = (data1.numberUnits * 100) + (data2.numberUnits * 100)
                spktrains_all = []
                for spktrain in neo_data.segments[0].spiketrains:
                    spktrains_all.append( spktrain - tstart*pq.ms )
                return spktrains_all
        else:
            return neo_data.segments[0].spiketrains
    
    # Private function for segregating respective spike trains for each cell assembly within a Neo object.
    def __split_spiketrains(self, dataname, neo_data, turnon):
        """Given the Neo object of a given data name ("basedata" or "propdata" or "reldata", see :py:meth:`.__create_datastructures`),
        this function returns a dictionary with "units" in the structured data as keys and its value the respective spike trains (all cell units in an assembly, i.e. the "unit").
        """
        ca_size = self.neural3assoc_topology.fsa.CA_SIZE
        #
        spkindices = lambda n : (0,ca_size) if (n==0) else ( (n*ca_size), (n*ca_size)+ca_size )
        #
        data = getattr(self, dataname) # "basedata" or "propdata" or "reldata"
        overallspikes = neo_data.segments[0].spiketrains
        #overallspikes = self.__get_overallspikes(dataname, neo_data, turnon) # technically w/o this spike train is not split
        #
        parsed_spiketrains = {"all": overallspikes} # this will be the returned value
        for unit in data.units:
            n = data.getUnitNumber(unit)
            indx = spkindices(n)
            parsed_spiketrains.update( {unit: overallspikes[ indx[0] : indx[-1] ] } )
        return parsed_spiketrains
    
    # Plotting functions
    # Private function to get first key for self.results dictionary
    def __get_resultskey(self, dataname):
        """Returns key string for ``self.results``. This is because the attribute names for the data structures are a shortened
        version of the keys in ``self.results`` with the exception of ``self.basedata``."""
        if dataname=="reldata":
            return "relation" # corresponding key in self.results is "relation"
        elif dataname=="propdata":
            return "property" # corresponding key in self.results is "property"
        else:
            return dataname.strip("data") # basedata -> base
    
    # Private function for each subplot in :py:meth:`.plot_all`
    def __subplot_all(self, dataname, colorname, subplotobject):
        """For a given ``subplotobject`` this function returns matplotlib's `eventplot <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.eventplot.html>`_
        for spiketrains from all the cell units of all the cell assemblies for a given data structure (assigned by the ``dataname`` argument)."""
        legpatches = []
        data = getattr(self, dataname) # "basedata" or "propdata" or "reldata"
        #clrs = self.__generate_compatible_subplot_colors(colorname, data.numberUnits)
        clrs = cm.get_cmap(colorname, 12) # "Reds", "Blues", "Greens"
        for unit in data.units:
            i = data.getUnitNumber(unit)
            #if self.test_metadata[0]=="all":
            subplotobject.eventplot( self.results[ self.__get_resultskey(dataname) ][ unit ],
                                     color = clrs(1.0 - (i*0.1) ) )
            legpatches.append( mpatches.Patch(color=clrs( 1.0-(i*0.1) ), label=unit) )
            #else:
            #    subplotobject.eventplot( self.results[ self.__get_resultskey(dataname) ][ unit ],
            #                             color = clrs[i] )
            #    legpatches.append( mpatches.Patch(color=clrs[i], label=unit) )
        subplotobject.legend( handles=legpatches, shadow=True )
        
    # Private function for returning color map (a tuple) or list of colors
    def __generate_compatible_subplot_colors(self, colorname, numberUnits):
        """This is no longer used but left here for potential/future use."""
        if self.test_metadata[0]=="all": # returns a tuple
            return cm.get_cmap(colorname, 12) # "Reds", "Greens", "Blues"
        else:
            return ['C{}'.format(i) for i in range(numberUnits)]
    
    def plot_all(self, form="1"):
        """Plotting spike trains from all the units within every cell assembly can be done in two ways.
        
        **Default**
        
        Plots all the spike trains in one plot.
        
        **``form = "2"``**
        
        Plot of three subplots such that each subplot (invoking :py:meth:`.__subplot_all`) is the matplotlib's `eventplot <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.eventplot.html>`_
        of `SpikeTrain <https://neo.readthedocs.io/en/stable/api_reference.html#neo.core.SpikeTrain>`_ such that
        
        +---------+------------------------------------------------------------+----------------------------------+
        | subplot | spike trains for all cell units in all cell assemblies for | color map for ``turnon = "all"`` |
        +=========+============================================================+==================================+
        | top     | subject (a.k.a base in ``basedata``)                       | Reds                             |
        +---------+------------------------------------------------------------+----------------------------------+
        | middle  | object (a.k.a property in ``propdata``)                    | Blues                            |
        +---------+------------------------------------------------------------+----------------------------------+
        | bottom  | predicate (a.k.a relation in ``reldata``)                  | Greens                           |
        +---------+------------------------------------------------------------+----------------------------------+
        
        """
        plt.close() # close any previous figure
        if form=="1":
            datanames = ["basedata", "propdata", "reldata"]
            colornames = ["Reds", "Blues", "Greens"]
            legpatches = []
            for i in range(len(datanames)):
                dataname = datanames[i]
                data = getattr(self, dataname)
                clrs = cm.get_cmap(colornames[i], 12)
                for unit in data.units:
                    j = data.getUnitNumber(unit)
                    plt.eventplot( self.results[ self.__get_resultskey(dataname) ][ unit ],
                                   color = clrs(1.0 - (i*0.1) ) )
                    legpatches.append( mpatches.Patch(color=clrs( 1.0-(i*0.1) ), label=unit) )
            plt.legend( handles=legpatches, shadow=True, loc='upper right' )
            plt.ylabel("cell units\nper CA")
            plt.xlabel("time (ms)")
            plt.title("Three-way Association")
        elif form=="2":
            fig, ((sp1),
                  (sp2),
                  (sp3)) = plt.subplots(3,1, sharex=True)
            # If turnon is "all" Red color for subject, i.e. base Therefore its spectrum is used here
            self.__subplot_all("basedata", "Reds", sp1)
            #
            # If turnon is "all" Blue color for object, i.e. property
            self.__subplot_all("propdata", "Blues", sp2)
            #
            # If turnon is "all" Green color for predicate, i.e. relation
            self.__subplot_all("reldata", "Greens", sp3)
            #
            sp1.title.set_text("Subject (base)")
            sp2.title.set_text("Object (property)")
            sp3.title.set_text("Predicate (relation)")
            #
            sp1.set(ylabel="cell units\nper CA")
            sp2.set(ylabel="cell units\nper CA")
            sp3.set(ylabel="cell units\nper CA")
            sp3.set(xlabel="time (ms)")
            #
            plt.subplots_adjust( hspace=0.5 ) # spacing for the each subplot title
        plt.show()
        
    def plot_specific(self, basename=None, relname=None, propname=None):
        """Plot of three subplots such that each is the matplotlib's `eventplot <https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.eventplot.html>`_
        of `SpikeTrain <https://neo.readthedocs.io/en/stable/api_reference.html#neo.core.SpikeTrain>`_ of all the cell units in cell assembly for given
        
        * ``basename`` in top subplot for subject, color is red
        * ``relname`` in middle subplot for predicate, color is green
        * ``propname`` in bottom subplot for object, color is blue
        
        """
        #plt.close() # close any previous figure
        fig, ((sp1),
              (sp2),
              (sp3)) = plt.subplots(3,1, sharex=True)
        #
        sp1.eventplot( self.results["base"][basename], color="red" ) # animal
        z_patch = mpatches.Patch( color="red", label=basename )
        sp1.legend( handles=[z_patch], shadow=True )
        #
        sp2.eventplot( self.results["property"][propname], color="blue" ) # food
        z_patch = mpatches.Patch( color="blue", label=propname )
        sp2.legend( handles=[z_patch], shadow=True )
        #
        sp3.eventplot( self.results["relation"][relname], color="green" ) # eats
        z_patch = mpatches.Patch( color="green", label=relname )
        sp3.legend( handles=[z_patch], shadow=True, loc='upper left' )
        #
        sp1.title.set_text('Subject (base)')
        sp2.title.set_text('Object (property)')
        sp3.title.set_text('Predicate (relation)')
        #
        sp1.set(ylabel="cell units\nper CA")
        sp2.set(ylabel="cell units\nper CA")
        sp3.set(ylabel="cell units\nper CA")
        sp3.set(xlabel="time (ms)")
        #
        plt.subplots_adjust( hspace=.5 ) # spacing for the each subplot title
        plt.show()


# ~/nmcog/spinnaker/associate/nealassoc/make3Assoc.py
#
# Documentation by Lungsi 24 March 2020
#
# Code based on Jan2020 Chris Huyck et al.
# http://www.cwa.mdx.ac.uk/NEAL/code/assocMemJan2020.tar.gz
#
# This creates a three way association topology.  One of the types
# is the base type (typically a node in a semantic net).  The other
# two are the same but one is named relation and the other property.
# If you turn two on, the third should come on.  You should also be
# able to get the (base) to spread up the inheritance hierarchy.
#
# Testing is weak.  It has only been tested in nest.

from nmcog.spinnaker.specialfunction.neal import FSAHelperFunctions
from .makeInheritanceHier import NeuralInheritanceClass

# added for nmcog
import spynnaker8 as sim
from nmcog.spinnaker.specialfunction.neal import NealCoverFunctions

class NeuralThreeAssocClass:
    """
    
    +---------------------------------------+------------------------------------------------+----------------------------------------------+
    | Methods                               |  Argument                                      | Caller                                       |
    +=======================================+================================================+==============================================+
    | :py:meth:`.createBaseNet`             | - instance of :ref:`InheritanceReaderClass`    | :ref:`NEAL3Way`                              |
    +---------------------------------------+------------------------------------------------+----------------------------------------------+
    | :py:meth:`.createAssociationTopology` | - instance of :ref:`UnitReaderClass`           | :ref:`NEAL3Way`                              |
    |                                       | - different instance of :ref:`UnitReaderClass` |                                              |
    +---------------------------------------+------------------------------------------------+----------------------------------------------+
    | :py:meth:`.addAssociations`           | - a structured data                            | :ref:`NEAL3Way`                              |
    +---------------------------------------+------------------------------------------------+----------------------------------------------+
    
    
    **Note:**
    
    * Specifically the instance of :ref:`InheritanceReaderClass` is the base data for :ref:`NEAL3Way`. An example base data is
    
    ::
    
        bases = {"units": ["animal", "mammal", "bird", "canary"],
                 "is-a": [ ["canary", "bird"], ["bird", "animal"], ["mammal", "animal"] ]}
        associate = {"properties": ["food", "fur", "flying", "yellow"], # properties to be associated between base units and its relations
                     "relations": ["eats", "likes", "travels", "has", "colored"], # relations associated with properties and base units
                     "connections": [ ["animal", "eats", "food"], ["mammal", "has", "fur"], # specific combos of base-props-relations
                                      ["bird", "travels", "flying"], ["canary", "colored", "yellow"]] }
    
    
    """
    #class variables
    numPropertyCAs = -1
    numRelationshipCAs = -1
    neuralHierarchyTopology = None
    baseStructure = None
    propStructure = None
    relStructure = None

    #def __init__(self, simName,sim,neal,spinnVersion,fsa):
    def __init__(self, simName="spinnaker", spinnVersion=8): # modified for nmcog
        self.simName = simName
        self.sim = sim
        self.spinnVersion = spinnVersion
        self.neal = NealCoverFunctions()
        self.fsa = FSAHelperFunctions()

    def createNeurons(self,numPropertyNeurons,numRelationNeurons):
        """Creates a two population (``propertyCells`` and ``relationCells``) of `IF_cond_exp <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_ neurons with :ref:`FSAHelperFunctions` ``.CELL_PARAMS``."""
        self.propertyCells = self.sim.Population(numPropertyNeurons,
                    self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.relationCells = self.sim.Population(numRelationNeurons,
                    self.sim.IF_cond_exp, self.fsa.CELL_PARAMS)

    def setRecord(self):
        """Records spikes of all the neurons in the ``propertyCells`` and ``relationCells`` populations created via :py:meth:`.createNeurons`."""
        self.propertyCells.record(['spikes'])
        self.relationCells.record(['spikes'])

    #Make a binary CA for each unit
    def makeCAs(self):
        """Makes ``numCAs`` assemblies of cell, i.e. connect neurons in the two population created via :py:meth:`.createNeurons`.
        
        * ``numPropertyCAs`` assemblies of ``propertyCells``
        * ``numRelationCAs`` assemblies of ``relationCells``
        
        Like the "hierarchy topology" created with :ref:`NeuralInheritanceClass` ``.createNeuralInheritanceHierarchy`` the practical meaning of the cell assemblies of "property" and "relations" in terms of extracting/visualizing the neuron units in a population is as follows.
        
        * For our example of four property units ``["food", "fur", "flying", "yellow"]`` there will be four cell assemblies for "property".
        * For our example of four relation units ``["eats", "likes", "travels", "has", "colored"]`` there will be five cell assemblies for "relation".
        * By default :ref:`FSAHelperFunctions` ``.CA_SIZE`` = 10, which is the number of neuron units in each assembly.
        
            - This is done by invoking :py:meth:`createNeurons`
            - The result is ``self.propertyCells`` and ``self.relationCells``
            
        * After simulating for a particular runtime, spikes from all the neurons (i.e. all neuron units in all the neuronal populations in all the assemblies) by
        
        ::
        
            allpropertyspikes = self.propertyCells.get_data( variables=["spikes"] )
            allrelationspikes = self.relationCells.get_data( variables=["spikes"] )
        
        
            - It should be noted that the above objects are `Neo Blocks <https://neo.readthedocs.io/en/latest/api_reference.html#neo.core.Block>`_
            - It is a Neo Block with only one `Segment <https://neo.readthedocs.io/en/latest/api_reference.html#neo.core.Segment>`_
        
        * Therefore,
        
            - Spike trains for all the neuron units in the cell assembly for the property units "food", "fur", "flying", and "yellow" are respectively
            
                * ``allpropertyspikes.segments[0].spiketrains[0]`` for "food"
                * ``allpropertyspikes.segments[0].spiketrains[1]`` for "fur"
                * ``allpropertyspikes.segments[0].spiketrains[2]`` for "flying"
                * ``allpropertyspikes.segments[0].spiketrains[3]`` for "yellow"
            
            - Similarly, for the relation units "eats", "likes", "travels", "has", and "colored" are respectively
            
                * ``allrelationspikes.segments[0].spiketrains[0]`` for "eats"
                * ``allrelationspikes.segments[0].spiketrains[1]`` for "likes"
                * ``allrelationspikes.segments[0].spiketrains[2]`` for "travels"
                * ``allrelationspikes.segments[0].spiketrains[3]`` for "has"
                * ``allrelationspikes.segments[0].spiketrains[4]`` for "colored"
        
        """
        for CA in range (0,self.numPropertyCAs):
            self.fsa.makeCA(self.propertyCells,CA)
        for CA in range (0,self.numRelationCAs):
            self.fsa.makeCA(self.relationCells,CA)
    
    #main function
    def createBaseNet(self,baseNodeStructure):
        """Creates a hierarchy topology for base data. See :ref:`NeuralInheritanceClass.createNeuralInheritanceHierarchy` to know more about hierarchy topology."""
        self.baseStructure = baseNodeStructure
        #self.neuralHierarchyTopology = NeuralInheritanceClass(self.simName,
        #        self.sim,self.neal,self.spinnVersion,self.fsa)
        self.neuralHierarchyTopology = NeuralInheritanceClass() # modified for nmcog
        self.neuralHierarchyTopology.createNeuralInheritanceHierarchy(baseNodeStructure)

    def createAssociationTopology(self,propertyStructure,relationStructure):
        """Given two instances of :ref:`UnitReaderClass` this method creates cell assemblies (:py:meth:`.makeCAs`) based on the two neuron populations (:py:meth:`.createNeurons`)."""
        self.propStructure = propertyStructure
        self.relStructure = relationStructure
        self.numPropertyCAs = propertyStructure.numberUnits 
        numberPropertyNeurons = self.numPropertyCAs * self.fsa.CA_SIZE
        self.numRelationCAs = relationStructure.numberUnits 
        numberRelationNeurons = self.numRelationCAs * self.fsa.CA_SIZE
        self.createNeurons(numberPropertyNeurons,numberRelationNeurons)
        self.setRecord()
        self.makeCAs()

    #Add synapses that make a 2/3 CA.
    def addThreeAssoc(self,assocTuple):
        """Given an association tuple (list of three strings representing base, relation, and property) this method connects
        
        * The hierarchy topology to the cell assemblies for property.
        * The hierarchy topology to the cell assemblies for relation.
        * The cell assemblies for property to the hierarchy topology.
        * The cell assemblies for property to the cell assemblies for relation.
        * The cell assemblies for relation to the hierarchy topology.
        * The cell assemblies for relation to the cell assemblies for property.
        
        """
        base = assocTuple[0]
        relation = assocTuple[1]
        property = assocTuple[2]
        baseNum = self.baseStructure.getUnitNumber(base)
        propertyNum = self.propStructure.getUnitNumber(property)
        relationNum = self.relStructure.getUnitNumber(relation)
        #print(assocTuple)
        #print(type(base),  base)
        self.fsa.stateHalfTurnsOnState(self.neuralHierarchyTopology.cells,
                                       baseNum,self.propertyCells,propertyNum)
        self.fsa.stateHalfTurnsOnState(self.neuralHierarchyTopology.cells,
                                       baseNum,self.relationCells,relationNum)

        self.fsa.stateHalfTurnsOnState(self.propertyCells,propertyNum,
                                self.neuralHierarchyTopology.cells,baseNum)
        self.fsa.stateHalfTurnsOnState(self.propertyCells,propertyNum,
                                       self.relationCells,relationNum)

        self.fsa.stateHalfTurnsOnState(self.relationCells,relationNum,
                                self.neuralHierarchyTopology.cells,baseNum)
        self.fsa.stateHalfTurnsOnState(self.relationCells,relationNum,
                                       self.propertyCells,propertyNum)

    def addAssociations(self,assocStructure):
        """Add associations for each tuple in a data structure of associations by calling :py:meth:`.addThreeAssoc` for each association."""
        #print(assocStructure.numberAssocs)
        for assocNum in range (0,assocStructure.numberAssocs):
            self.addThreeAssoc(assocStructure.assocs[assocNum])


    #-print function
    #def printSpikes(self):
    #    """."""
    #    self.propertyCells.write_data("testProps.pkl")
    #    self.relationCells.write_data("testRels.pkl")
    #    self.neuralHierarchyTopology.cells.write_data("test3.pkl")


    def printPklSpikes(self,inFileName,outFileName):
        """Legacy."""
        outFileHandle = open(outFileName,'w')
        inFileHandle = open(inFileName)
        neoObj = pickle.load(inFileHandle)
        segments = neoObj.segments
        segment = segments[0]
        spikeTrains = segment.spiketrains
        neurons = len(spikeTrains)
        for neuronNum in range (0,neurons):
            if (len(spikeTrains[neuronNum])>0):
                spikes = spikeTrains[neuronNum]
                for spike in range (0,len(spikes)):
                    #outFileHandle.print(neuronNum, spikes[spike])
                    outString = str(neuronNum) + " " + str(spikes[spike]) +"\n";
                    outFileHandle.write(outString)
        inFileHandle.close()
        outFileHandle.flush()
        outFileHandle.close()

    def printSpikes(self,fileName):
        """."""
        # Commented out for nmcog
        #if ((self.simName =="spinnaker") and (self.spinnVersion == 7)):
        #    suffix = ".sp"
        #elif ((self.simName =="nest") or
        #      ((self.simName =="spinnaker") and (self.spinnVersion == 8))):
        #    suffix = ".pkl"
        suffix = ".pkl"
        basePklFile = "results/"+fileName +"Bases" + suffix
        #baseSpFile = "results/"+fileName +"Bases.sp"
        self.neuralHierarchyTopology.cells.write_data(basePklFile)
        #self.printPklSpikes(basePklFile,baseSpFile)
        propPklFile = "results/"+fileName +"Props"+ suffix
        self.propertyCells.write_data(propPklFile)
        relPklFile = "results/"+fileName +"Rels"+ suffix
        #self.relationCells.printSpikes(relPklFile)
        self.relationCells.write_data(relPklFile)

    #--test functions
    #Set up spike generators to start each unit, then stop them.
    #When run, each unit should persist, but only that unit 
    #should.
    def makeGenerator(self,genTime):
        """Set up spike generator to start each unit. ``genTime`` is a float representing the start time."""
        genTimes = genTime
        genTimeArray = {'spike_times': [genTimes]}
        spikeGen=self.sim.Population(1,self.sim.SpikeSourceArray,genTimeArray)
        return spikeGen

    #now stop all of the units after each individual unit is tested.
    def createStopAll(self,stopTimes,cells,numUnits):
        """Stop all the spikes generated, **each** generated using :py:meth:`.makeGenerator`."""
        stopTimeArray = {'spike_times': [stopTimes]}
        stopSpikeGen=self.sim.Population(1,self.sim.SpikeSourceArray,
                                         stopTimeArray)
        for unit in range (0,numUnits): 
            self.fsa.turnOffStateFromSpikeSource(stopSpikeGen,cells,unit)

    def createSimpleTest(self):
        """Legacy."""
        primeTimes = [5]
        primeArray = {'spike_times': [primeTimes]}
        generator=sim.Population(1,sim.SpikeSourceArray,primeArray)
        
        self.fsa.turnOnStateFromSpikeSource(generator,self.propertyCells,1)
        self.fsa.turnOnStateFromSpikeSource(generator,self.relationCells,2)
        self.neuralHierarchyTopology.createSimpleTest()

    #call this with 0-3 items to be stimulated (but typically two)
    def createTwoTest(self,baseNum,propNum,relNum):
        """Legacy."""
        primeTimes = [5]
        primeArray = {'spike_times': [primeTimes]}
        generator=self.sim.Population(1,self.sim.SpikeSourceArray,primeArray)
        
        if (baseNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,
                        self.neuralHierarchyTopology.cells,baseNum)
        if (propNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,self.propertyCells,
                                                propNum)
        if (relNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,self.relationCells,
                                                relNum)

    def createTestPrimeAllBaseUnits(self,firstTestStart,hierarchy):
        """."""
        primeWeight = 0.014#0.017
        oneTestDuration = 100.0
        timeBetweenSteps = 5.0
        numberPrimeSteps = 10
        for primeEpoch in range (0,hierarchy.numCAs): 
            startTime = 25.0 + (primeEpoch*oneTestDuration)+firstTestStart
            primeTimes = []
            for primeStep in range (0,numberPrimeSteps): 
                primeTimes = primeTimes + [startTime + 
                                           (timeBetweenSteps*primeStep)]
            generator = self.makeGenerator(primeTimes)

            for unit in range (0,hierarchy.numCAs): #numCAs is numUnits
                self.fsa.stimulateStateFromSpikeSource(generator,hierarchy.cells,
                                                       unit,primeWeight)
        return firstTestStart + (hierarchy.numCAs*oneTestDuration)

    #call this with 0-3 items to be stimulated (but typically two)
    def createTwoPrimeTest(self,baseNum,propNum,relNum):
        """."""
        primeTimes = [5]
        primeArray = {'spike_times': [primeTimes]}
        generator=self.sim.Population(1,self.sim.SpikeSourceArray,primeArray)
        
        if (baseNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,
                        self.neuralHierarchyTopology.cells,baseNum)
            self.createTestPrimeAllBaseUnits(5.0,self.neuralHierarchyTopology)

        if (propNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,self.propertyCells,
                                                propNum)
        if (relNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,self.relationCells,
                                                relNum)

    #call this with 0-3 items to be stimulated (but typically two) 
    def createTwoPrimeTestPoisson(self,baseNum,propNum,relNum):
        """Legacy."""
        PoissonParameters = {'duration': 200.0, 'start': 0.0, 'rate': 200.0} # rate = 200 Hz  
        generator=self.sim.Population(1,self.sim.SpikeSourcePoisson,PoissonParameters)
        
        if (baseNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,
                        self.neuralHierarchyTopology.cells,baseNum)
            self.createTestPrimeAllBaseUnits(5.0,self.neuralHierarchyTopology)

        if (propNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,self.propertyCells,
                                                propNum)
        if (relNum >= 0):
            self.fsa.turnOnStateFromSpikeSource(generator,self.relationCells,
                                                relNum)

    def createTestAllUnits(self,firstTestStart,cells,numUnits):
        """Given a start time ``firstTestStart``, the state for ``numUnits`` cell assemblies of a neuron population (``cells``) is turned on.
        The method returns the time after turning on the state for the last cell assembly.
        """
        #print(numUnits)
        oneTestDuration = 100.0
        stopTimes = []
        for unit in range (0,numUnits): 
            startTime = 25.0 + (unit*oneTestDuration)+firstTestStart
            generator = self.makeGenerator([startTime])
            self.fsa.turnOnStateFromSpikeSource(generator,cells,unit)
            lastTime =(((unit+1)*oneTestDuration)+firstTestStart)
            stopTimes = stopTimes + [lastTime]

        #print(stopTimes)
        self.createStopAll(stopTimes,cells,numUnits)
        return lastTime

    def createUnitTests(self):
        """Turns on the state for every cell assembly in
        
        * The hierarchy topology (base data)
        * The populations for ``propertyCells`` and ``relationCells``.
        
        """
        baseUnitTime = self.neuralHierarchyTopology.createTestAllUnits(0.0)
        propUnitTime = self.createTestAllUnits(baseUnitTime,self.propertyCells,
                                               self.numPropertyCAs)
        relUnitTime = self.createTestAllUnits(propUnitTime,self.relationCells,
                                               self.numRelationCAs)
        #print(baseUnitTime,propUnitTime,relUnitTime)
        return relUnitTime



from distutils.core import setup

with open( "README.md", "r" ) as fh:
    long_description = fh.read()

setup(
  name = "nmcog",
  packages = [
      "nmcog",
      "nmcog.spinnaker",
      "nmcog.spinnaker.discriminate",
      "nmcog.spinnaker.associate",
      "nmcog.spinnaker.associate.nealassoc",
      "nmcog.spinnaker.cognitivemap",
      "nmcog.spinnaker.cognitivemap.nealmentalmap",
      #"nmcog.spinnaker.cognitiveplanning",
      #"nmcog.spinnaker.cognitiveplanning.nealplanning",
      "nmcog.spinnaker.specialfunction.neal",
      ],
  version = "0.2.27",
  license="BSD Clause-3",
  description = "Module (Library) for simulation cognitive models in Neuromorphic hardwares.",
  long_description = long_description,
  long_description_content_type = "text/markdown",
  author = "Lungsi Ngwua",
  author_email = "lungsi.ngwua@cnrs.fr",
  url = "https://github.com/myHBPwork/NMCog",
  download_url = "https://github.com/NMCog/archive/v0.2.27.tar.gz",
  keywords = ["NEUROMORPHIC", "COGNITIVE", "SPIKING", "NEURAL NETWORK", "SPINNAKER", "BRAINSCALES"],
  #install_requires=[
  #        "numpy",
  #        "quantities",
  #        "neo",
  #        "spynnaker8",
  #    ],
  classifiers=[
    "Development Status :: 3 - Alpha",      # Choose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    "Intended Audience :: Developers",      # Define that your audience are developers
    "Topic :: Software Development :: Build Tools",
    "License :: OSI Approved :: BSD License",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
  ],
  python_requires = ">=3.5",
)



# ~/nmcog/spinnaker/cognitivemap/nealmentalmap/placecellsysClass.py
#
# Documentation by Lungsi 6 April 2020
#
#This is driven by an FSA that manages the binding and the retrievals.
#Initially, (startState) it waits for a binding event.  It then
#tryToBind, which uses a timer since it takes time to bind.  If it
#succeeds it goes to bindOnState
#bindOn turns on the bindingTimer
#the bindingTimer turns on bindDone state
#bindDone -> start via bindDoneTimer.
#
#Fail states occurs with bindFailState twoPlacesFact, twoObjectsFact,
#and notEnoughPlaceObjectsFact.
#
#retrievePlaceState is turned on externally by a query.
#It turns off start.  It allows the binding nets to fire, and when a place
#fires, it turns on the placeRetrievedFact (extra neurons in the automaton net).
#Together these turn on retrievePlaceDone, which then goes back to startState
#
#A similar mechanism is used for retrieveObject.
#

from nmcog.spinnaker.specialfunction.neal import NealCoverFunctions
from nmcog.spinnaker.specialfunction.neal import TimerClass

# added for nmcog
import spynnaker8 as sim
from nmcog.spinnaker.specialfunction.neal import FSAHelperFunctions

class PlaceCellSystemClass:
    """
    
    +----------------------------------+-------------------------------+
    | Methods                          | Argument                      |
    +==================================+===============================+
    | :py:meth:`.createObjects`        | integer ``numberObjects``     |
    +----------------------------------+-------------------------------+
    | :py:meth:`.createPlaces`         | integer ``numberPlaces``      |
    +----------------------------------+-------------------------------+
    | :py:meth:`.createAutomaton`      | - integer ``tryToBindStates`` |
    |                                  | - integer ``bindingStates``   |
    |                                  | - integer ``bindDoneStates``  |
    +----------------------------------+-------------------------------+
    | :py:meth:`.connectObjects`       | -                             |
    +----------------------------------+-------------------------------+
    | :py:meth:`.connectPlaces`        | -                             |
    +----------------------------------+-------------------------------+
    | :py:meth:`.connectAutomaton`     | -                             |
    +----------------------------------+-------------------------------+
    | :py:meth:`.setupCogMapRecording` | -                             |
    +----------------------------------+-------------------------------+
    | :py:meth:`.makeLearningSynapses` | -                             |
    +----------------------------------+-------------------------------+
    
    
    +---------------+-----------------+--------------------------------+
    | Methods       | Cell assemblies | Available cells (for analysis) |
    +===============+=================+================================+
    | An object     | to bind cells   | - ``self.objectBindCells``     |
    |               | to bind on      | - ``self.objectBindOnCells``   |
    |               | to bind done    | - ``self.objectBindDoneCells`` |
    |               | to query on     | - ``self.queryOnObjectCells``  |
    |               | to answer       | - ``self.answerObjectCells``   |
    +---------------+-----------------+--------------------------------+
    | A place       | to bind cells   | - ``self.placeBindCells``      |
    |               | to bind on      | - ``self.placeBindOnCells``    |
    |               | to bind done    | - ``self.placeBindDoneCells``  |
    |               | to query on     | - ``self.queryOnPlaceCells``   |
    |               | to answer       | - ``self.answerPlaceCells``    |
    +---------------+-----------------+--------------------------------+
    | The Automaton |      15         | - ``self.automatonCells``      |
    +---------------+-----------------+--------------------------------+

    
    
    +-------------------------------+-------------------+-----------------------------------------------------+
    | Names of 15 Automaton states  | State number      | Methods where state is used                         |
    +===============================+===================+=====================================================+
    | ``startState``                | 0                 | - :py:meth:`.connectStartState`                     |
    |                               |                   | - :py:meth:`.connectBindFailState`                  |
    |                               |                   | - :py:meth:`.connectBindDoneState`                  |
    |                               |                   | - :py:meth:`.connectRetrieveObjectState`            |
    |                               |                   | - :py:meth:`.connectRetrieveObjectDoneState`        |
    |                               |                   | - :py:meth:`.connectRetrievePlaceState`             |
    |                               |                   | - :py:meth:`.connectRetrievePlaceDoneState`         |
    |                               |                   | - :py:meth:`.sourceStartsAutomaton`                 |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``tryBindState``              | 1                 | - :py:meth:`.connectStartState`                     |
    |                               |                   | - :py:meth:`.connectBindOnState`                    |
    |                               |                   | - :py:meth:`.connectBindFailState`                  |
    |                               |                   | - :py:meth:`.sourceTurnsOnBind`                     |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``bindOnState``               | 2                 | - :py:meth:`.connectBindOnState`                    |
    |                               |                   | - :py:meth:`.connectBindFailState`                  |
    |                               |                   | - :py:meth:`.bindOnPrimesPlaceAndObjectBind`        |
    |                               |                   | - :py:meth:`.connectBindDoneState`                  |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``bindFailState``             | 3                 | - :py:meth:`.connectStartState`                     |
    |                               |                   | - :py:meth:`.connectBindFailState`                  |
    |                               |                   | - :py:meth:`.bindFailStopsPlaceObjectOn`            |
    |                               |                   | - :py:meth:`.bindFailStopsPlaceObjectOn`            |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``bindDoneState``             | 4                 | - :py:meth:`.connectStartState`                     |
    |                               |                   | - :py:meth:`.connectBindDoneState`                  |
    |                               |                   | - :py:meth:`.connectBindDoneState`                  |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``retrieveObjectState``       | 5                 | - :py:meth:`.retrieveObjectPrimesPlaceBind`         |
    |                               |                   | - :py:meth:`.connectRetrieveObjectState`            |
    |                               |                   | - :py:meth:`.connectRetrieveObjectDoneState`        |
    |                               |                   | - :py:meth:`.sourceTurnOnRetrieveObjectFromPlace`   |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``retrievePlaceState``        | 6                 | - :py:meth:`.retrievePlacePrimesObjectBind`         |
    |                               |                   | - :py:meth:`.connectRetrievePlaceState`             |
    |                               |                   | - :py:meth:`.connectRetrievePlaceDoneState`         |
    |                               |                   | - :py:meth:`.sourceTurnOnRetrievePlaceFromObject`   |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``retrieveObjectDoneState``   | 7                 | - :py:meth:`.connectStartState`                     |
    |                               |                   | - :py:meth:`.connectRetrieveObjectDoneState`        |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``retrievePlaceDoneState``    | 8                 | - :py:meth:`.connectStartState`                     |
    |                               |                   | - :py:meth:`.connectRetrievePlaceDoneState`         |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``lastState``                 | 8                 |                                                     |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``onePlaceOneObjectFact``     | ``lastState`` + 1 | - :py:meth:`.setupObjectPlaceFacts`                 |
    |                               |                   | - :py:meth:`.connectBindOnState`                    |
    |                               |                   | - :py:meth:`.connectBindFailState`                  |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``twoPlacesFact``             | ``lastState`` + 2 | - :py:meth:`.setupObjectPlaceFacts`                 |
    |                               |                   | - :py:meth:`.connectBindFailState`                  |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``twoObjectsFact``            | ``lastState`` + 3 | - :py:meth:`.setupObjectPlaceFacts`                 |
    |                               |                   | - :py:meth:`.connectBindFailState`                  |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``notEnoughPlaceObjectsFact`` | ``lastState`` + 4 | - :py:meth:`.connectBindFailState`                  |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``placeRetrievedFact``        | ``lastState`` + 5 | - :py:meth:`.answerPlaceStartsPlaceRetrievedFact`   |
    |                               |                   | - :py:meth:`.connectRetrievePlaceDoneState`         |
    +-------------------------------+-------------------+-----------------------------------------------------+
    | ``objectRetrievedFact``       | ``lastState`` + 6 | - :py:meth:`.answerObjectStartsObjectRetrievedFact` |
    |                               |                   | - :py:meth:`.connectRetrieveObjectDoneState`        |
    +-------------------------------+-------------------+-----------------------------------------------------+
    
    
    """
    numberAutomatonStates = 15
    numberObjects = -1
    numberPlaces = -1
    
    #def __init__(self,simName,sim,neal,spinnVersion,fsa):
    def __init__(self,simName="spinnaker",spinnVersion=8):
        self.sim = sim
        self.neal = NealCoverFunctions()
        self.fsa = FSAHelperFunctions()
        # added for nmcog
        self.tryToBindTimer = TimerClass(simName = simName,
                                         spinnVersion = spinnVersion)
        self.bindingTimer = TimerClass(simName = simName,
                                       spinnVersion = spinnVersion)
        self.bindDoneTimer = TimerClass(simName = simName,
                                        spinnVersion = spinnVersion)
        # this instantiates the three timers but the timercells are not created here
        # they are created in self.createTimers where default values are used for the
        # number of states but user can pass custom values.
        # however, self.createTimers is called by self.createAutomaton therefore
        # these custom values must be passed into this function which takes effect on
        # the self.createTimers function.
        #
        #self.automatonCells = self.createAutomaton() # commented out for nmcog

    #-----FSA functions and constants
    #state names for the Automaton
    startState = 0
    tryBindState = 1
    bindOnState = 2
    bindFailState = 3
    bindDoneState = 4
    retrieveObjectState = 5
    retrievePlaceState = 6
    retrieveObjectDoneState = 7
    retrievePlaceDoneState = 8
    lastState = 8
    onePlaceOneObjectFact = lastState+1 #one placeBindOn and one objectBindOn on
    twoPlacesFact = lastState+2 #two or more placeBindOn
    twoObjectsFact = lastState+3 #two or more objectBindOn
    notEnoughPlaceObjectsFact = lastState+4 #0 or 1 objects place in total
    placeRetrievedFact = lastState+5
    objectRetrievedFact = lastState+6

    #Each object and place half connects to the onePlaceOneObjectFact
    #Note that 2 of either will turn it on as will more than one of each.
    #These are handled elsewhere to convert to the bindFailState
    def setupObjectPlaceFacts(self):
        """
        This function connects objects and places with the automaton (see :py:meth:`.createAutomaton`).
        
        Below show the connection for one object and one place. But note that the connections are made
        for all objects and all places with the automaton.

        ::
        
            wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww
            w                      An Object                      w
          xxwxxxxxxxxxxxxxxxxxxxxxxxxxxx                          w
          x w                          x                          w
          x w   oooooooooooooo   oooooooooooooo   oooooooooooooo  w
          x w   o    binds   o   o   binds    o   o   binds    o  w
          x w   o    cells   o   o  on-cells  o   o done-cells o  w
          x w   oooooooooooooo   oooooooooooooo   oooooooooooooo  w
          x w                                                     w
          x w   oooooooooooooo   oooooooooooooo                   w
          x w   o  inquire   o   o   answer   o                   w
          x w   o  on-cells  o   o about cell o                   w
          x w   oooooooooooooo   oooooooooooooo                   w
          x w                                                     w
          x w                                                     w
          x wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww
          x 
          x 
          x 
          x aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          x a                               Automaton                                     a
          x a                                                                             a
          x a  oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo  a
          x a  o startState o   o tryBindState o   o bindFailState o   o bindDoneState o  a
          x a  oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo  a
          xxaxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx                                           a
            a  ooooooooooooooooooooooooooo x  x oooooooooooooooooooooooooo                a
            a  o retrieveObjectDoneState o x  x o retrievePlaceDoneState o                a
            a  ooooooooooooooooooooooooooo x  x oooooooooooooooooooooooooo                a
            a                <1/2-ON> Vxxxxx  V <1/2-ON>                                  a
            a  ooooooooooooooooooooooooo   ooooooooooooooo                                a
            a  o onePlaceOneObjectFact o   o bindOnState o                                a
            a  ooooooooooooooooooooooooo   ooooooooooooooo                                a
            a                         ^       ^                                           a
            a                <1/2-ON> x       x <1/2-ON>                                  a
            a                         xxxxxxxxx                                           a
            a                                 x                                           a
            aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
                                              x
                                              x
                                              x         
            ppppppppppppppppppppppppppppppppppppppppppppppppppppppp
            p                       A Place   x                   p
            p                                 x                   p
            p   oooooooooooooo   oooooooooooooo   oooooooooooooo  p
            p   o    binds   o   o   binds    o   o   binds    o  p
            p   o    cells   o   o  on-cells  o   o done-cells o  p
            p   oooooooooooooo   oooooooooooooo   oooooooooooooo  p
            p                                                     p
            p   oooooooooooooo   oooooooooooooo                   p
            p   o  inquire   o   o   answer   o                   p
            p   o  on-cells  o   o about cell o                   p
            p   oooooooooooooo   oooooooooooooo                   p
            p                                                     p
            p                                                     p
            ppppppppppppppppppppppppppppppppppppppppppppppppppppppp
        
        **NOTE:**
        
        * All the states of an automaton are **not** shown in the above illustration.
        * See :ref:`FSAHelperFunctions` ``.stateHalfTurnsOnState``
        
        """
        #print(self.numberObjects)
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateHalfTurnsOnState(self.objectBindOnCells,objectNumber,
                            self.automatonCells,self.onePlaceOneObjectFact)
            self.fsa.stateHalfTurnsOnState(self.objectBindOnCells,objectNumber,
                            self.automatonCells,self.twoObjectsFact)

        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateHalfTurnsOnState(self.placeBindOnCells,placeNumber,
                            self.automatonCells,self.onePlaceOneObjectFact)
            self.fsa.stateHalfTurnsOnState(self.placeBindOnCells,placeNumber,
                            self.automatonCells,self.twoPlacesFact)

    #---Make the fsa that governs the map.  
    def stateStopsPlaceDone(self,stateNumber):
        """
        The given state of the automaton (see :py:meth:`.createAutomaton`) turns-off the
        ``self.placeBindDoneCells`` for all the places, i.e. over all ``self.numberPlaces``.
        
        Below shows the ``startState`` of the automaton turning-off the state represented by
        ``self.placeBindDoneCells`` of just one place. Note that this will be done for all the
        places in ``self.numberPlaces``.

        ::
        
            aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
            a                               Automaton                                     a
          xxaxxxxxxxx                                                                     a
          x a       x                                                                     a
          x a  oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo  a
          x a  o startState o   o tryBindState o   o bindFailState o   o bindDoneState o  a
          x a  oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo  a
          x a                                                                             a
          x a  ooooooooooooooooooooooooooo   oooooooooooooooooooooooooo                   a
          x a  o retrieveObjectDoneState o   o retrievePlaceDoneState o                   a
          x a  ooooooooooooooooooooooooooo   oooooooooooooooooooooooooo                   a
          x a                                                                             a
          x a                                                                             a
          x a                                                                             a
          x aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          x
          xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx <OFF>
                                                        x            
            ppppppppppppppppppppppppppppppppppppppppppppppppppppppp
            p                       A Place             x         p
            p                                           V         p
            p   oooooooooooooo   oooooooooooooo   oooooooooooooo  p
            p   o    binds   o   o   binds    o   o   binds    o  p
            p   o    cells   o   o  on-cells  o   o done-cells o  p
            p   oooooooooooooo   oooooooooooooo   oooooooooooooo  p
            p                                                     p
            p   oooooooooooooo   oooooooooooooo                   p
            p   o  inquire   o   o   answer   o                   p
            p   o  on-cells  o   o about cell o                   p
            p   oooooooooooooo   oooooooooooooo                   p
            p                                                     p
            p                                                     p
            ppppppppppppppppppppppppppppppppppppppppppppppppppppppp
        
        **NOTE:**
        
        * All the states of an automaton are **not** shown in the above illustration.
        * See :ref:`FSAHelperFunctions` ``.stateTurnsOffState``
        
        """
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.placeBindDoneCells,placeNumber)

    def stateStopsObjectDone(self,stateNumber):
        """See :py:meth:`.stateStopsPlaceDone` but replace place for object."""
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.objectBindDoneCells,objectNumber)

    def stateStopsDone(self,stateNumber):
        """See :py:meth:`.stateStopsPlaceDone` and :py:meth:`.stateStopsObjectDone`"""
        self.stateStopsPlaceDone(stateNumber)
        self.stateStopsObjectDone(stateNumber)

    def connectStartState(self):
        """
        Following the creation of ``self.automatonCells`` (:py:meth:`.createAutomaton`)
        this function connects the states of the automaton.
        
        Connection for an automaton is done as follows
        
        ::
        
                        <OFF>
                    xxxxxxxxxxxxxx
                    x   <1/2-ON> x
                    x    xxxxxx  x         xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
                    \/   x   \/  x         x                                            x
            oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo   x
            o startState o   o tryBindState o   o bindFailState o   o bindDoneState o   x
            oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo   x
            x   x     x  x         <OFF>        /\                  /\                  x
            x   x     x  xxxxxxxxxxxxxxxxxxxxxxxxx       <OFF>       x                  x
            x   x     xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx                  x
            x   xxxxxxxxxxxxxxxxxxxxxxxxxxx                                             x
            \/<OFF>                       \/<OFF>                                       x
            ooooooooooooooooooooooooooo   oooooooooooooooooooooooooo                    x
            o retrieveObjectDoneState o   o retrievePlaceDoneState o                    x
            ooooooooooooooooooooooooooo   oooooooooooooooooooooooooo                    x
                                                                                        x
            xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            x                                                                   <OFF>
            x
            \/              tryToBindTimer
            ooooooooooooo   ooooooooooooo   ooooooooooooo
            o  state-0  o   o state-... o   o  state-k  o
            ooooooooooooo   ooooooooooooo   ooooooooooooo
                
        Note that for the above illustration to be complete two more steps are required
        
        * see :py:meth:`.stateStopsDone`
        * see :py:meth:`.statePreventsOutput`
        * All the states of an automaton are **not** shown in the above illustration.
        
        Also see
        
        * :ref:`FSAHelperFunctions` ``.stateHalfTurnsOnState``
        * :ref:`FSAHelperFunctions` ``.stateTurnsOffState``
        * :ref:`TimerClass` ``.stateStartsTimer``
        
        """
        #the bind signal moves from start to tryBind
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.startState,
                                       self.automatonCells,self.tryBindState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.tryBindState,
                                       self.automatonCells,self.startState)
        self.tryToBindTimer.stateStartsTimer(self.automatonCells,
                                                 self.tryBindState)

        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                                       self.automatonCells,self.bindFailState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                                       self.automatonCells,self.bindDoneState)

        self.stateStopsDone(self.startState)

        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                            self.automatonCells,self.retrieveObjectDoneState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                            self.automatonCells,self.retrievePlaceDoneState)

        self.statePreventsOutput(self.startState)


    def statePreventsPlaceOutput(self,stateNumber):
        """See :py:meth:`.statePreventsObjectOutput` but replace object for place."""
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.answerPlaceCells,placeNumber)

    def statePreventsObjectOutput(self,stateNumber):
        """
        The given state of the automaton (see :py:meth:`.createAutomaton`) turns-off the
        ``self.answerObjectCells`` for all the places, i.e. over all ``self.numberPlaces``.
        
        Below shows the ``startState`` of the automaton turning-off the state represented by
        ``self.answerObjectCells`` of just one place. Note that this will be done for all the
        places in ``self.numberPlaces``.

        ::
        
            aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
            a                               Automaton                                     a
          xxaxxxxxxxx                                                                     a
          x a       x                                                                     a
          x a  oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo  a
          x a  o startState o   o tryBindState o   o bindFailState o   o bindDoneState o  a
          x a  oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo  a
          x a                                                                             a
          x a  ooooooooooooooooooooooooooo   oooooooooooooooooooooooooo                   a
          x a  o retrieveObjectDoneState o   o retrievePlaceDoneState o                   a
          x a  ooooooooooooooooooooooooooo   oooooooooooooooooooooooooo                   a
          x a                                                                             a
          x a                                                                             a
          x a                                                                             a
          x aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          x
          xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
                                                                       x           
            wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww    x
            w                      An Object                      w    x
            w                                                     w    x
            w   oooooooooooooo   oooooooooooooo   oooooooooooooo  w    x
            w   o    binds   o   o   binds    o   o   binds    o  w    x
            w   o    cells   o   o  on-cells  o   o done-cells o  w    x
            w   oooooooooooooo   oooooooooooooo   oooooooooooooo  w    x
            w                                                     w    x
            w   oooooooooooooo   oooooooooooooo                   w    x
            w   o  inquire   o   o   answer   o <xxxxxxxxxxxxxxxxxpxxxxx  <OFF>
            w   o  on-cells  o   o about cell o                   w
            w   oooooooooooooo   oooooooooooooo                   w
            w                                                     w
            w                                                     w
            wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww
        
        **NOTE:**
        
        * All the states of an automaton are **not** shown in the above illustration.
        * See :ref:`FSAHelperFunctions` ``.stateTurnsOffState``
        
        """
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.answerObjectCells,objectNumber)

    def statePreventsOutput(self,stateNumber):
        """See :py:meth:`.statePreventsPlaceOutput` and :py:meth:`.statePreventsObjectOutput`"""
        self.statePreventsPlaceOutput(stateNumber)
        self.statePreventsObjectOutput(stateNumber)

    #-----Connect States----
    def connectBindOnState(self):
        """
        The given state of the automaton (see :py:meth:`.createAutomaton`) turns-off the
        ``self.answerObjectCells`` for all the places, i.e. over all ``self.numberPlaces``.
        
        Below shows the ``startState`` of the automaton turning-off the state represented by
        ``self.answerObjectCells`` of just one place. Note that this will be done for all the
        places in ``self.numberPlaces``.

        ::
        
            aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
            a                               Automaton                                     a
            a                                                                             a
            a  oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo  a
            a  o startState o   o tryBindState o   o bindFailState o   o bindDoneState o  a
            a  oooooooooooooo   oooooooooooooooo   ooooooooooooooooo   ooooooooooooooooo  a
            a                              x  ^ <OFF>                                     a
            a  ooooooooooooooooooooooooooo x  x oooooooooooooooooooooooooo                a
            a  o retrieveObjectDoneState o x  x o retrievePlaceDoneState o                a
            a  ooooooooooooooooooooooooooo x  x oooooooooooooooooooooooooo                a
            a                     <1/2-ON> V  x                                           a
            a  ooooooooooooooooooooooooo   ooooooooooooooo                                a
            a  o onePlaceOneObjectFact o   o bindOnState o                                a
            a  ooooooooooooooooooooooooo   ooooooooooooooo                                a
            a                    ^   x        ^  x   ;;;;;                                a
            a              <OFF> x   x <STIM> x  x   ;;;;;                                a
            a                    x   xxxxxxxxxx  x   ;;;;;                                a
            a                    xxxxxxxxxxxxxxxxx   ;;;;;                                a
            aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
                                                     ;;;;;
                                                     ;;;;;
                                                   ..;;;;;..
                                                    ':::::'
                                                      ':`
            ttttttttttttttttttttttttttttttttttttttttttttttttttt
            t               tryToBindTimer                    t
            t  ooooooooooooo   ooooooooooooo   ooooooooooooo  t
            t  o  state-0  o   o state-... o   o  state-k  o  t
            t  ooooooooooooo   ooooooooooooo   ooooooooooooo  t
            t                                                 t
            ttttttttttttttttttttttttttttttttttttttttttttttttttt
        
        Note that for the above illustration to be complete the automaton must be connected to
        
        * all the objects and places as shown in :py:meth:`.setupObjectPlaceFacts`
        * see :py:meth:`.statePreventsOutput` but with ``self.bindOnState`` of the automaton, unlike the example in :py:meth:`.statePreventsOutput`
        
        Also,
        
        * All the states of an automaton are **not** shown in the above illustration.
        * See :ref:`FSAHelperFunctions` ``.stateHalfTurnsOnState``
        * See :ref:`FSAHelperFunctions` ``.stateStimulatesState``
        * See :ref:`FSAHelperFunctions` ``.stateTurnsOffState``
        * See :ref:`TimerClass` ``.stateStopsTimer``
        """
        #1p1o fact moves from tryBind to bindOn
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.tryBindState,
                                       self.automatonCells,self.bindOnState)
        self.setupObjectPlaceFacts()
        #Half_on is 0.0012, here we go slower so that it comes up slower
        #if two places is on
        #funnyWt = 0.001
        #if ((self.neal.simulator == 'spinnaker') and (self.neal.spinnVersion == 8)):
        #    funnyWt = 0.0011
        funnyWt = 0.0011

        self.fsa.stateStimulatesState(self.automatonCells,
                        self.onePlaceOneObjectFact,self.automatonCells,
                                      self.bindOnState,funnyWt)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindOnState,
                                       self.automatonCells,self.tryBindState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindOnState,
                                self.automatonCells,self.onePlaceOneObjectFact)
        #stop tryBindTimer
        self.tryToBindTimer.stateStopsTimer(self.automatonCells,
                                            self.bindOnState)

        self.statePreventsOutput(self.bindOnState)
        
    def connectBindFailState(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        #2p, 2o, or timer goes to bindFail 
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.tryBindState,
                                       self.automatonCells,self.bindFailState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.twoPlacesFact,
                                       self.automatonCells,self.bindFailState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.twoObjectsFact,
                                       self.automatonCells,self.bindFailState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.notEnoughPlaceObjectsFact,self.automatonCells,
                                       self.bindFailState)
        self.tryToBindTimer.timerStartsState(self.automatonCells, #timeout fail
                                            self.notEnoughPlaceObjectsFact)

        #fail prevents on
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                                       self.automatonCells,self.bindOnState)

        #turn states and timers off
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                                       self.automatonCells,self.tryBindState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                                       self.automatonCells,self.twoPlacesFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                                       self.automatonCells,self.twoObjectsFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                            self.automatonCells,self.onePlaceOneObjectFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                            self.automatonCells,self.notEnoughPlaceObjectsFact)
        self.tryToBindTimer.stateStopsTimer(self.automatonCells,
                                            self.bindFailState)

        #fail goes right back to start
        self.fsa.stateTurnsOnState(self.automatonCells,self.bindFailState,
                            self.automatonCells,self.startState)

    def bindFailStopsPlaceObjectOn(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                            self.placeBindOnCells,placeNumber)
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                            self.objectBindOnCells,objectNumber)

    def bindOnPrimesPlaceAndObjectBind(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateHalfTurnsOnState(self.automatonCells,self.bindOnState,
                            self.placeBindCells,placeNumber)
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateHalfTurnsOnState(self.automatonCells,self.bindOnState,
                            self.objectBindCells,objectNumber)

    def stopPlaceAndObjectBind(self,stateNumber):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.placeBindCells,placeNumber)
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.objectBindCells,objectNumber)

    def connectBindDoneState(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        #timer moves to binddone
        self.bindingTimer.stateStartsTimer(self.automatonCells, 
                                            self.bindOnState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.bindOnState,
                                       self.automatonCells,self.bindDoneState)
        self.bindingTimer.timerStartsState(self.automatonCells, 
                                            self.bindDoneState)

        self.bindOnPrimesPlaceAndObjectBind()

        self.fsa.stateTurnsOffState(self.automatonCells,self.bindDoneState,
                            self.automatonCells,self.bindOnState)
        self.stopPlaceAndObjectBind(self.bindDoneState)

        #binding done returns to start via timer
        self.bindDoneTimer.stateStartsTimer(self.automatonCells, 
                                            self.bindDoneState)
        self.bindDoneTimer.timerStartsState(self.automatonCells, 
                                            self.startState)

        self.statePreventsOutput(self.bindDoneState)


    def retrieveObjectPrimesPlaceBind(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateHalfTurnsOnState(self.automatonCells,
                    self.retrieveObjectState,self.placeBindCells,placeNumber)

    def answerPlaceStartsPlaceRetrievedFact(self):
        """."""
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOnState(self.answerPlaceCells,placeNumber,
                    self.automatonCells,self.placeRetrievedFact)

    def connectRetrieveObjectState(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.startState,
                        self.automatonCells,self.retrieveObjectState)
        self.retrieveObjectPrimesPlaceBind()
        self.statePreventsPlaceOutput(self.retrieveObjectState)
        self.answerPlaceStartsPlaceRetrievedFact()
        self.fsa.stateTurnsOffState(self.automatonCells,
                self.retrieveObjectState,self.automatonCells,self.startState)

    def connectRetrieveObjectDoneState(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.retrieveObjectState,self.automatonCells,
                                       self.retrieveObjectDoneState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.objectRetrievedFact,self.automatonCells,
                                       self.retrieveObjectDoneState)
        self.fsa.stateTurnsOffState(self.automatonCells,
            self.retrieveObjectDoneState,self.automatonCells,
                                       self.retrieveObjectState)
        #don't need the output on anymore
        self.statePreventsObjectOutput(self.retrieveObjectDoneState)
        self.stopPlaceAndObjectBind(self.retrieveObjectDoneState)

        self.fsa.stateTurnsOnState(self.automatonCells,
            self.retrieveObjectDoneState,self.automatonCells,
                                       self.startState)

    def retrievePlacePrimesObjectBind(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateHalfTurnsOnState(self.automatonCells,
                    self.retrievePlaceState,self.objectBindCells,objectNumber)

    def answerObjectStartsObjectRetrievedFact(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOnState(self.answerObjectCells,objectNumber,
                    self.automatonCells,self.objectRetrievedFact)

    def connectRetrievePlaceState(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.startState,
                        self.automatonCells,self.retrievePlaceState)
        self.retrievePlacePrimesObjectBind()
        self.statePreventsObjectOutput(self.retrievePlaceState)
        self.answerObjectStartsObjectRetrievedFact()
        self.fsa.stateTurnsOffState(self.automatonCells,
                self.retrievePlaceState,self.automatonCells,self.startState)

    def connectRetrievePlaceDoneState(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.retrievePlaceState,self.automatonCells,
                                       self.retrievePlaceDoneState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.placeRetrievedFact,self.automatonCells,
                                       self.retrievePlaceDoneState)
        self.fsa.stateTurnsOffState(self.automatonCells,
            self.retrievePlaceDoneState,self.automatonCells,
                                       self.retrievePlaceState)
        #don't need the output on anymore
        self.statePreventsPlaceOutput(self.retrievePlaceDoneState)
        self.stopPlaceAndObjectBind(self.retrievePlaceDoneState)

        self.fsa.stateTurnsOnState(self.automatonCells,
            self.retrievePlaceDoneState,self.automatonCells,
                                       self.startState)

    def connectAutomaton(self):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.connectStartState()
        self.connectBindOnState()
        self.connectBindFailState()
        self.bindFailStopsPlaceObjectOn()
        self.connectBindDoneState()
        self.connectRetrieveObjectState()
        self.connectRetrievePlaceState()
        self.connectRetrieveObjectDoneState()
        self.connectRetrievePlaceDoneState()

    #def createTimers(self):
        #self.tryToBindTimer = TimerClass(self.simName,self.sim,self.neal,
        #                                 self.spinnVersion,self.fsa,10)
        #self.bindingTimer = TimerClass(self.simName,self.sim,self.neal,
        #                               self.spinnVersion,self.fsa,10)
        #self.bindDoneTimer = TimerClass(self.simName,self.sim,self.neal,
        #                                self.spinnVersion,self.fsa,6)
    def createTimers(self, tryToBindStates=10, bindingStates=10, bindDoneStates=6):
        """
        Given the arguments ``tryToBindStates``, ``bindingStates``, and ``bindDoneStates`` this function creates three timers,
        
        * ``self.tryToBindTimer`` for its given number of states ``tryToBindStates``
        
            - a timer for accounting the time taken for a binding event
        
        * ``self.bindingTimer`` for its given number of states ``bindingStates``
        
            - a timer for binding
        
        * ``self.bindDoneTimer`` for its given number of states ``bindDoneStates``
        
            - a timer for when the binding is achieved
        
        A timer is a collection of cell assemblies such that
        
        * the number of cell assemblies for a timer correspond to the number of possible states of the timer
        * its cell assemblies are connected using :ref:`TimerClass` ``.makeStopNoRestartTimerSynapses``
        
            - the timer goes from state to state and the last stops on its own
            - if the state that turns it ON remains ON, this timer will not restart until its done
        
        Regardless of the state, i.e. the cell assembly, the
        
        * ``self.tryToBindTimer.timerCells``,
        * ``self.bindingTimer.timerCells``, and
        * ``self.bindDoneTimer.timerCells``
        
        are its `neuron population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ for the above timers.
        
        """
        # modified for nmcog
        self.tryToBindTimer.createNeurons( tryToBindStates ) # although default values are set
        self.bindingTimer.createNeurons( bindingStates )     # user can pass custom number of states
        self.bindDoneTimer.createNeurons( bindDoneStates )
        #the timerClass sets recording
        self.tryToBindTimer.makeStopNoRestartTimerSynapses()
        self.bindingTimer.makeStopNoRestartTimerSynapses()
        self.bindDoneTimer.makeStopNoRestartTimerSynapses()

    #def createAutomaton(self): # commented out for nmcog
    def createAutomaton(self, tryToBindStates=10, bindingStates=10, bindDoneStates=6):
        """
        Given the arguments ``tryToBindStates``, ``bindingStates``, and ``bindDoneStates`` this function creates,
        
        * cell assemblies, and
        * three timers (see :py:meth:`.createTimers`)
        
        The cell assemblies are created using :ref:`FSAHelperFunctions` ``.makeCA``. An automaton with five states will therefore have five corresponding cell assemblies as shown below
        
        ::
        
            ooooooooooooo   ooooooooooooo   ooooooooooooo   ooooooooooooo   ooooooooooooo
            o Automaton o   o Automaton o   o Automaton o   o Automaton o   o Automaton o
            o  state-0  o   o  state-1  o   o  state-2  o   o  state-3  o   o  state-4  o
            ooooooooooooo   ooooooooooooo   ooooooooooooo   ooooooooooooo   ooooooooooooo
            
        * the number of cell assemblies for *an* automaton is determined by ``self.numberAutomatonStates`` (default = 15)
        * the number of `neuron populations <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ within a cell assembly is given by :ref:`FSAHelperFunctions` ``.CA_SIZE`` (default = 10)
        * the number of neuron units within a population is ``CA_SIZE * numberAutomatonStates`` (default = 150)
        
        This function returns the automaton cells, the prototype `neuron populations <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_ used in constructing the above cell assemblies.
        
        """
        numNeurons = self.fsa.CA_SIZE * self.numberAutomatonStates
        
        cells=self.sim.Population(numNeurons,self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        for stateNumber in range (0,self.numberAutomatonStates):
            self.fsa.makeCA(cells,stateNumber)

        #self.createTimers() # commented out for nmcog
        self.createTimers(tryToBindStates=tryToBindStates, bindingStates=bindingStates,
                          bindDoneStates=bindDoneStates)
        #return cells               # replaced for nmcog
        self.automatonCells = cells # with this (i.e, createAutomaton() no longer called in __init__()

    def createObjects(self, numberObjects):
        """For a given ``numberObjects`` this function creates five cell assemblies using :ref:`FSAHelperFunctions` ``.makeCA``
        
        * functionally, one cell assembly each for
        
            - binding the cells
            - binding the on-cells ("on" state of the cells)
            - binding the done-cells ("done" state of the cells)
            - inquiring about the on-cells
            - answering about the cells
        
        * all cell assemblies
        
            - have ``numberObjects`` times :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource`` cells in a `population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_
            - uses the same cell type, `IF_cond_exp <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#pyNN.standardmodels.cells.IF_cond_exp>`_
        
        Therefore each object is represented by five cell assemblies.
        
        ::
    
            oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo
            o   CA to bind   o   o   CA to bind   o   o   CA to bind   o
            o object's cells o   o  the on-cells  o   o the done-cells o
            oooooooooooooooooo   oooooooooooooooooo   oooooooooooooooooo
            
            oooooooooooooooooo   oooooooooooooooooo
            o CA to inquire  o   o  CA to answer  o
            o    on-cells    o   o about the cell o
            oooooooooooooooooo   oooooooooooooooooo

        For a population of :math:`n \\times 10` cells the cell assembly comprises of ten such population.
        
        ::
        
            oooooooooooooooooooooooooooooooooooooooooooooooo
            o                                              o
            o   .-.      .-.      .-.      .-.      .-.    o
            o  ( 0 )    ( 2 )    ( 4 )    ( 6 )    ( 8 )   o
            o   '-'      '-'      '-'      '-'      '-'    o
            o                                              o
            o        cell assembly of CA_SIZE = 10         o
            o                                              o
            o   .-.      .-.      .-.      .-.      .-.    o
            o  ( 1 )    ( 3 )    ( 5 )    ( 7 )    ( 9 )   o
            o   '-'      '-'      '-'      '-'      '-'    o
            o                                              o
            oooooooooooooooooooooooooooooooooooooooooooooooo
        
        **NOTE:**
        
        * The value 10 is the default value for :ref:`FSAHelperFunctions` ``.CA_SIZE`
        * See :ref:`FSAHelperFunctions` ``.getCAConnectors` for how the ten populations are connected.
        
        """
        self.numberObjects = numberObjects
        #print 'objects',numberObjects
        numberObjectCells = self.numberObjects*self.fsa.CA_SIZE
        self.objectBindCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.objectBindOnCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.objectBindDoneCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.queryOnObjectCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.answerObjectCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)

        for objectNumber in range (0,self.numberObjects):
            self.fsa.makeCA(self.objectBindCells,objectNumber)
            self.fsa.makeCA(self.objectBindOnCells,objectNumber)
            self.fsa.makeCA(self.objectBindDoneCells,objectNumber)
            self.fsa.makeCA(self.queryOnObjectCells,objectNumber)
            self.fsa.makeCA(self.answerObjectCells,objectNumber)

    def connectObjects(self):
        """Connects the objects created using :py:meth:`.createObjects`
        
        Connection for an object is done as follows
        
        ::
            
                            <1/2-ON>           <OFF>
            xxxxxxxxx    xxxxxxxxxxxxxx   xxxxxxxxxxxxxxxx
            x       x    V            x   V              x
            x   oooooooooooooo   oooooooooooooo   oooooooooooooo
            x   o    binds   o   o   binds    o   o   binds    o
            x   o    cells   o   o  on-cells  o   o done-cells o
            x   oooooooooooooo   oooooooooooooo   oooooooooooooo
            x       ^    x           <ON>               ^
            x       x    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            x       x <1/2-ON>
            x   oooooooooooooo   oooooooooooooo
            x   o  inquire   o   o   answer   o
            x   o  on-cells  o   o about cell o
            x   oooooooooooooo   oooooooooooooo
            x                           ^
            x          <ON>             x
            xxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        
        
        * for <1/2-ON> see :ref:`FSAHelperFunctions` ``.stateHalfTurnsOnState``
        * for <ON> see :ref:`FSAHelperFunctions` ``.stateTurnsOnState``
        * for <OFF> see :ref:`FSAHelperFunctions` ``.stateTurnsOffState``
        
        """
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateHalfTurnsOnState(self.objectBindOnCells,objectNumber,
                                       self.objectBindCells,objectNumber)
            self.fsa.stateTurnsOnState(self.objectBindCells,objectNumber,
                                       self.objectBindDoneCells,objectNumber)
            self.fsa.stateTurnsOffState(self.objectBindDoneCells,objectNumber,
                                       self.objectBindOnCells,objectNumber)
            self.fsa.stateHalfTurnsOnState(self.queryOnObjectCells,objectNumber,
                                       self.objectBindCells,objectNumber)
            self.fsa.stateTurnsOnState(self.objectBindCells,objectNumber,
                                       self.answerObjectCells,objectNumber)

    #Create neurons for each place and make them CAs.
    def createPlaces(self, numberPlaces):
        """See :py:meth:`.createObjects`"""
        self.numberPlaces = numberPlaces
        #print('places',numberPlaces)
        numberPlaceCells = self.numberPlaces*self.fsa.CA_SIZE
        self.placeBindCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.placeBindOnCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.placeBindDoneCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.queryOnPlaceCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.answerPlaceCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)

        for placeNumber in range (0,self.numberPlaces):
            self.fsa.makeCA(self.placeBindCells,placeNumber)
            self.fsa.makeCA(self.placeBindOnCells,placeNumber)
            self.fsa.makeCA(self.placeBindDoneCells,placeNumber)
            self.fsa.makeCA(self.queryOnPlaceCells,placeNumber)
            self.fsa.makeCA(self.answerPlaceCells,placeNumber)

    def connectPlaces(self):
        """See :py:meth:`.connectObjects`"""
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateHalfTurnsOnState(self.placeBindOnCells,placeNumber,
                                       self.placeBindCells,placeNumber)
            self.fsa.stateTurnsOnState(self.placeBindCells,placeNumber,
                                       self.placeBindDoneCells,placeNumber)
            self.fsa.stateTurnsOffState(self.placeBindDoneCells,placeNumber,
                                       self.placeBindOnCells,placeNumber)
            self.fsa.stateHalfTurnsOnState(self.queryOnPlaceCells,placeNumber,
                                       self.placeBindCells,placeNumber)
            self.fsa.stateTurnsOnState(self.placeBindCells,placeNumber,
                                       self.answerPlaceCells,placeNumber)

    def setupCogMapRecording(self):
        """Record spikes for
        
        +-----------------------------------------------+---------------------------------------------+
        | cell assemblies (i.e. state) within an Object | cell assemblies (i.e. state) within a Place |
        +===============================================+=============================================+
        | ``self.objectBindCells``                      | ``self.placeBindCells``                     |
        +-----------------------------------------------+---------------------------------------------+
        | ``self.objectBindOnCells``                    | ``self.placeBindOnCells``                   |
        +-----------------------------------------------+---------------------------------------------+
        | ``self.objectBindDoneCells``                  | ``self.placeBindDoneCells``                 |
        +-----------------------------------------------+---------------------------------------------+
        | ``self.queryOnObjectCells``                   | ``self.queryOnPlaceCells``                  |
        +-----------------------------------------------+---------------------------------------------+
        | ``self.answerObjectCells``                    | ``self.answerPlaceCells``                   |
        +-----------------------------------------------+---------------------------------------------+
        
        and also for ``self.automatonCells``
        
        """
        self.placeBindCells.record({'spikes'})
        self.objectBindCells.record({'spikes'})
        self.automatonCells.record({'spikes'})
        self.placeBindOnCells.record({'spikes'})
        self.objectBindOnCells.record({'spikes'})
        self.placeBindDoneCells.record({'spikes'})
        self.objectBindDoneCells.record({'spikes'})
        self.queryOnPlaceCells.record({'spikes'})
        self.queryOnObjectCells.record({'spikes'})
        self.answerPlaceCells.record({'spikes'})
        self.answerObjectCells.record({'spikes'})

    def getWellConnectedConn(self,fromSize,toSize,weight):
        """."""
        connector = []
        for fromNeuron in range (0,fromSize):
            for toNeuron in range (0, toSize):
                connector=connector+[(fromNeuron,toNeuron,weight,
                                      self.neal.DELAY)]
        return connector

    def makeLearningSynapses(self):
        """See :ref:`FSAHelperFunctions` ``.CA_SIZE``"""
        numberPlaceBindCells = self.numberPlaces*self.fsa.CA_SIZE
        numberObjectBindCells = self.numberObjects*self.fsa.CA_SIZE
        synWeight = 0.0

        placeToObjectConnector = self.getWellConnectedConn(
            numberPlaceBindCells,numberObjectBindCells,synWeight)
        placeToObjectConnList = self.sim.FromListConnector(
            placeToObjectConnector)

        objectToPlaceConnector = self.getWellConnectedConn(
            numberObjectBindCells,numberPlaceBindCells,synWeight)
        objectToPlaceConnList = self.sim.FromListConnector(
            objectToPlaceConnector)
        stdp_model = self.sim.STDPMechanism(
                timing_dependence = self.sim.SpikePairRule(tau_plus=16.0, 
                                tau_minus=3.0,A_plus=0.002, A_minus=0.0),
                #timing_dependence = self.sim.SpikePairRule(tau_plus=16.7, 
                #                tau_minus=3.7,A_plus=0.002, A_minus=0.0),
                                #tau_minus=3.7,A_plus=0.005, A_minus=0.0),#works
                                #tau_minus=3.7,A_plus=0.001, A_minus=0.0002),
                                #tau_minus=0.01,A_plus=0.005, A_minus=0.002),
                weight_dependence = self.sim.AdditiveWeightDependence(
                    w_min=0.0, w_max=0.004) 
                    #w_min=0.0, w_max=0.015) #.015 .004 .0015
                    #w_min=0.0, w_max=0.02) 
            )

        self.sim.Projection(self.placeBindCells,self.objectBindCells,
                            self.sim.AllToAllConnector(),
                            #placeToObjectConnList,
                            synapse_type=stdp_model)

        self.sim.Projection(self.objectBindCells,self.placeBindCells,
                            #objectToPlaceConnList,
                            self.sim.AllToAllConnector(),
                            synapse_type=stdp_model)
        #print 'bob', placeToObjectConnector
        #print objectToPlaceConnector

    #---Defs for turning things on and off with spike sources.
    def sourceTurnsOnPlace(self, placeNumber, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.turnOnStateFromSpikeSource(source,self.placeBindCells,
                                            placeNumber)
    def sourceTurnsOffPlace(self, placeNumber, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.turnOffStateFromSpikeSource(source,self.placeBindCells,
                                            placeNumber)

    def sourceTurnsOnObject(self, objectNumber, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.turnOnStateFromSpikeSource(source,self.objectBindCells,
                                            objectNumber)
    def sourceTurnsOffObject(self, objectNumber, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.turnOffStateFromSpikeSource(source,self.objectBindCells,
                                            objectNumber)

    def sourceTurnsOnBind(self, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.halfTurnOnStateFromSpikeSource(source,self.automatonCells,
                                                self.tryBindState)

    def sourceTurnsOnObjectOn(self, objectNumber, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.turnOnStateFromSpikeSource(source,self.objectBindOnCells,
                                            objectNumber)

    def sourceTurnsOnPlaceOn(self, placeNumber, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.turnOnStateFromSpikeSource(source,self.placeBindOnCells,
                                            placeNumber)

    def sourceStartsAutomaton(self, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.turnOnStateFromSpikeSource(source,self.automatonCells,
                                             self.startState)

    def sourceTurnOnRetrieveObjectFromPlace(self, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.halfTurnOnStateFromSpikeSource(source,self.automatonCells,
                                                self.retrieveObjectState)

    def sourceTurnOnRetrievePlaceFromObject(self, source):
        """See :ref:`FSAHelperFunctions` ``.halfTurnOnStateFromSpikeSource``"""
        self.fsa.halfTurnOnStateFromSpikeSource(source,self.automatonCells,
                                                self.retrievePlaceState)

    def sourceTurnsOnPlaceQuery(self, source,placeNumber):
        """See :ref:`FSAHelperFunctions` ``.turnOnStateFromSpikeSource``
        This is meant for place cells unlike :py:meth:`.sourceTurnsOnObjectQuery`"""
        self.fsa.turnOnStateFromSpikeSource(source,self.queryOnPlaceCells,
                                            placeNumber)

    def sourceTurnsOnObjectQuery(self, source,objectNumber):
        """Given a `spikeSource <http://neuralensemble.org/docs/PyNN/reference/neuronmodels.html#spike-sources>`_ ,
        this method turns **ON** the state of the object cells; the cells are collectively in the form of a
        `PyNN population <http://neuralensemble.org/docs/PyNN/reference/populations.html>`_.
        
        The object cells collectively is part of an object identified by its ``objectNumber``
        
        **NOTE:**
        
        * See :ref:`FSAHelperFunctions` ``.turnOnStateFromSpikeSource``
        * Althought
        * This is meant for objects unlike :py:meth:`.sourceTurnsOnPlaceQuery`
        
        """
        self.fsa.turnOnStateFromSpikeSource(source,self.queryOnObjectCells,
                                            objectNumber)

    def printCogMapNets(self):
        """Legacy."""
        suffix = '.pkl'
        self.placeBindCells.write_data('results/cmPlaceBind'+suffix)
        self.objectBindCells.write_data('results/cmObjectBind'+suffix)
        self.automatonCells.write_data('results/cmAutomaton'+suffix)
        self.placeBindOnCells.write_data('results/cmPlaceBindOn'+suffix)
        self.objectBindOnCells.write_data('results/cmObjectBindOn'+suffix)
        self.placeBindDoneCells.write_data('results/cmPlaceBindDone'+suffix)
        self.objectBindDoneCells.write_data('results/cmObjectBindDone'+suffix)
        self.queryOnPlaceCells.write_data('results/cmQueryOnPlace'+suffix)
        self.queryOnObjectCells.write_data('results/cmQueryOnObject'+suffix)
        self.answerPlaceCells.write_data('results/cmAnswerPlace'+suffix)
        self.answerObjectCells.write_data('results/cmAnswerObject'+suffix)
        self.tryToBindTimer.printResults('results/cmTryToBindTimer')
        self.bindingTimer.printResults('results/cmBindingTimer')
        self.bindDoneTimer.printResults('results/cmBindDoneTimer')
